use gejdr_core::models::accounts::User;
use gejdr_core::sqlx::PgPool;
use oauth2::{CsrfToken, PkceCodeVerifier, TokenResponse};
use poem::web::Data;
use poem::{session::Session, web::Form};
use poem_openapi::payload::{Json, PlainText};
use poem_openapi::{ApiResponse, Object, OpenApi};

use crate::oauth::{DiscordOauthProvider, OauthProvider};

use super::errors::ErrorResponse;
use super::ApiCategory;

type Token =
    oauth2::StandardTokenResponse<oauth2::EmptyExtraTokenFields, oauth2::basic::BasicTokenType>;

pub struct AuthApi;

#[derive(Debug, Object, Clone, Eq, PartialEq, serde::Deserialize)]
struct DiscordCallbackRequest {
    code: String,
    state: String,
}

impl DiscordCallbackRequest {
    pub fn check_token(&self, token: &CsrfToken) -> Result<(), LoginStatusResponse> {
        if *token.secret().to_string() == self.state {
            Ok(())
        } else {
            Err(LoginStatusResponse::TokenError(Json(ErrorResponse {
                code: 500,
                message: "OAuth token error".into(),
                details: Some(
                    "OAuth provider did not send a message that matches what was expected".into(),
                ),
            })))
        }
    }
}

#[derive(ApiResponse)]
enum LoginStatusResponse {
    #[oai(status = 201)]
    LoggedIn(Json<UserInfo>),
    #[oai(status = 201)]
    LoggedOut(
        #[oai(header = "Location")] String,
        #[oai(header = "Cache-Control")] String,
    ),
    #[oai(status = 301)]
    LoginRedirect(
        #[oai(header = "Location")] String,
        #[oai(header = "Cache-Control")] String,
    ),
    #[oai(status = 500)]
    TokenError(Json<ErrorResponse>),
    #[oai(status = 500)]
    DatabaseError(Json<ErrorResponse>),
    #[oai(status = 503)]
    DiscordError(Json<ErrorResponse>),
}

#[derive(Debug, Eq, PartialEq, serde::Serialize, Object)]
struct UserInfo {
    id: String,
    username: String,
    display_name: Option<String>,
    avatar: Option<String>,
}

impl From<User> for UserInfo {
    fn from(value: User) -> Self {
        Self {
            id: value.id,
            username: value.username,
            display_name: value.name,
            avatar: value.avatar,
        }
    }
}

#[derive(ApiResponse)]
enum UserInfoResponse {
    #[oai(status = 201)]
    UserInfo(Json<UserInfo>),
    #[oai(status = 401)]
    Unauthorized,
    #[oai(status = 500)]
    DatabaseError(Json<ErrorResponse>),
    #[oai(status = 503)]
    DiscordError(Json<ErrorResponse>),
}

impl From<UserInfoResponse> for LoginStatusResponse {
    fn from(value: UserInfoResponse) -> Self {
        match value {
            UserInfoResponse::UserInfo(json) => Self::LoggedIn(json),
            UserInfoResponse::Unauthorized => unimplemented!(),
            UserInfoResponse::DatabaseError(json) => Self::DatabaseError(json),
            UserInfoResponse::DiscordError(json) => Self::DiscordError(json),
        }
    }
}

#[derive(ApiResponse)]
enum CsrfResponse {
    #[oai(status = 201)]
    Token(PlainText<String>),
}

#[OpenApi(prefix_path = "/v1/api/auth", tag = "ApiCategory::Auth")]
impl AuthApi {
    async fn fetch_remote_user(
        pool: Data<&PgPool>,
        token: Token,
    ) -> Result<UserInfoResponse, UserInfoResponse> {
        crate::api_wrapper::discord::get_user_profile(token.access_token().secret())
            .await
            .map_err(|e| {
                tracing::event!(
                    target: "auth-discord",
                    tracing::Level::ERROR,
                    "Failed to communicate with Discord: {}",
                    e
                );
                UserInfoResponse::DiscordError(Json(e.into()))
            })?
            .refresh_in_database(&pool)
            .await
            .map(|user| UserInfoResponse::UserInfo(Json(user.into())))
            .map_err(|e| {
                tracing::event!(
                    target: "auth-discord",
                    tracing::Level::ERROR,
                    "Database error: {}",
                    e
                );
                UserInfoResponse::DatabaseError(Json(e.into()))
            })
    }

    #[oai(path = "/signin/discord", method = "get")]
    async fn signin_discord(
        &self,
        oauth: Data<&DiscordOauthProvider>,
        session: &Session,
    ) -> LoginStatusResponse {
        let (auth_url, csrf_token, pkce_verifier) = oauth.0.auth_and_csrf();
        session.set("csrf", csrf_token);
        session.set("pkce", pkce_verifier);
        tracing::event!(
            target: "auth-discord",
            tracing::Level::INFO,
            "Signin through Discord",
        );
        LoginStatusResponse::LoginRedirect(auth_url.to_string(), "no-cache".to_string())
    }

    #[oai(path = "/callback/discord", method = "get")]
    async fn callback_discord(
        &self,
        Form(auth_request): Form<DiscordCallbackRequest>,
        oauth: Data<&DiscordOauthProvider>,
        pool: Data<&PgPool>,
        session: &Session,
    ) -> Result<LoginStatusResponse, LoginStatusResponse> {
        tracing::event!(
            target: "auth-discord",
            tracing::Level::INFO,
            "Discord callback",
        );
        let csrf_token = session.get::<CsrfToken>("csrf").ok_or_else(|| {
            LoginStatusResponse::TokenError(Json(ErrorResponse {
                code: 500,
                message: "Cannot fetch csrf token from session".to_string(),
                ..Default::default()
            }))
        })?;
        auth_request.check_token(&csrf_token)?;
        let pkce_verifier = session.get::<PkceCodeVerifier>("pkce").ok_or_else(|| {
            LoginStatusResponse::TokenError(Json(ErrorResponse {
                code: 500,
                message: "Cannot fetch pkce verifier from session".to_string(),
                ..Default::default()
            }))
        })?;
        let token = oauth
            .token(auth_request.code, pkce_verifier)
            .await
            .map_err(|e| LoginStatusResponse::TokenError(Json(e.into())))?;
        session.set("token", token.clone());
        Self::fetch_remote_user(pool, token)
            .await
            .map(std::convert::Into::into)
            .map_err(std::convert::Into::into)
    }

    #[oai(path = "/csrf", method = "get")]
    async fn csrf(&self, token: &poem::web::CsrfToken) -> CsrfResponse {
        CsrfResponse::Token(PlainText(token.0.clone()))
    }

    #[oai(path = "/signout", method = "post")]
    async fn signout(&self, session: &Session) -> LoginStatusResponse {
        session.purge();
        LoginStatusResponse::LoggedOut("/".to_string(), "no-cache".to_string())
    }

    #[oai(path = "/me", method = "get")]
    async fn user_info(
        &self,
        session: &Session,
        pool: Data<&PgPool>,
    ) -> Result<UserInfoResponse, UserInfoResponse> {
        let token = session
            .get::<Token>("token")
            .ok_or(UserInfoResponse::Unauthorized)?;
        Self::fetch_remote_user(pool, token).await
    }
}
