use poem_openapi::Object;
use reqwest::Error as ReqwestError;

use crate::api_wrapper::ApiError as ApiWrapperError;
use crate::errors::ApiError;

#[derive(Debug, serde::Serialize, Default, Object, PartialEq, Eq)]
pub struct ErrorResponse {
    pub code: u16,
    pub message: String,
    pub details: Option<String>,
}

impl From<ApiError> for ErrorResponse {
    fn from(value: ApiError) -> Self {
        match value {
            ApiError::Sql(e) => Self {
                code: 500,
                message: "SQL error".into(),
                details: Some(e.to_string()),
            },
            ApiError::TokenError(e) => Self {
                code: 500,
                message: "OAuth token error".into(),
                details: Some(e),
            },
            ApiError::Unauthorized => Self {
                code: 401,
                message: "Unauthorized!".into(),
                ..Default::default()
            },
            ApiError::OptionError => Self {
                code: 500,
                message: "Attempted to get a value, but none found".into(),
                ..Default::default()
            },
        }
    }
}

impl From<ReqwestError> for ErrorResponse {
    fn from(value: ReqwestError) -> Self {
        Self {
            code: 503,
            message: "Failed to communicate with Discord".into(),
            details: Some(value.status().map_or_else(
                || "Communication failed before we could hear back from Discord".into(),
                |status| format!("Discord sent back the error code {status}"),
            )),
        }
    }
}

impl From<ApiWrapperError> for ErrorResponse {
    fn from(source: ApiWrapperError) -> Self {
        match source {
            ApiWrapperError::Reqwest(e) => e.into(),
            ApiWrapperError::Api(e) => Self {
                code: if e.message.as_str().starts_with("401") {
                    401
                } else {
                    e.code
                },
                message: e.message,
                details: None,
            },
        }
    }
}

impl From<gejdr_core::sqlx::Error> for ErrorResponse {
    fn from(_value: gejdr_core::sqlx::Error) -> Self {
        Self {
            code: 500,
            message: "Internal database error".into(),
            ..Default::default()
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::api_wrapper::{ApiError as ApiWrapperError, DiscordErrorResponse};

    #[test]
    fn conversion_from_sql_api_error_works() {
        let sql_error = ApiError::Sql(gejdr_core::sqlx::Error::ColumnNotFound(
            "COLUMN_NAME".to_string(),
        ));
        let final_error = ErrorResponse::from(sql_error);
        let expected_error = ErrorResponse {
            code: 500,
            message: "SQL error".into(),
            details: Some("no column found for name: COLUMN_NAME".into()),
        };
        assert_eq!(expected_error, final_error);
    }

    #[test]
    fn conversion_from_token_error_works() {
        let initial_error = ApiError::TokenError("TOKEN ERROR".into());
        let final_error: ErrorResponse = initial_error.into();
        let expected_error = ErrorResponse {
            code: 500,
            message: "OAuth token error".into(),
            details: Some("TOKEN ERROR".into()),
        };
        assert_eq!(expected_error, final_error);
    }

    #[test]
    fn conversion_from_unauthorized_works() {
        let initial_error = ApiError::Unauthorized;
        let final_error: ErrorResponse = initial_error.into();
        let expected_error = ErrorResponse {
            code: 401,
            message: "Unauthorized!".into(),
            ..Default::default()
        };
        assert_eq!(expected_error, final_error);
    }

    #[test]
    fn conversion_from_option_error_works() {
        let initial_error = ApiError::OptionError;
        let final_error: ErrorResponse = initial_error.into();
        let expected_error = ErrorResponse {
            code: 500,
            message: "Attempted to get a value, but none found".into(),
            ..Default::default()
        };
        assert_eq!(expected_error, final_error);
    }

    #[tokio::test]
    async fn conversion_from_reqwest_error() {
        let err = reqwest::get("https://example.example/401").await;
        assert!(err.is_err());
        let expected = ErrorResponse {
            code: 503,
            message: "Failed to communicate with Discord".into(),
            details: Some("Communication failed before we could hear back from Discord".into()),
        };
        let actual: ErrorResponse = err.err().unwrap().into();
        assert_eq!(expected, actual);
    }

    #[tokio::test]
    async fn conversion_from_apiwrappererror_with_reqwest_error() {
        let err = reqwest::get("https://example.example/401").await;
        assert!(err.is_err());
        let err = ApiWrapperError::Reqwest(err.err().unwrap());
        let expected = ErrorResponse {
            code: 503,
            message: "Failed to communicate with Discord".into(),
            details: Some("Communication failed before we could hear back from Discord".into()),
        };
        let actual: ErrorResponse = err.into();
        assert_eq!(expected, actual);
    }

    #[test]
    fn conversion_from_apiwrappererror_with_401_discord_error() {
        let err = ApiWrapperError::Api(DiscordErrorResponse {
            code: 0,
            message: "401: Unauthorized".into(),
        });
        let expected = ErrorResponse {
            code: 401,
            message: "401: Unauthorized".into(),
            ..Default::default()
        };
        let actual: ErrorResponse = err.into();
        assert_eq!(expected, actual);
    }

    #[test]
    fn conversion_from_apiwrappererror_with_generic_discord_error() {
        let err = ApiWrapperError::Api(DiscordErrorResponse {
            code: 0,
            message: "Something else".into(),
        });
        let expected = ErrorResponse {
            code: 0,
            message: "Something else".into(),
            ..Default::default()
        };
        let actual: ErrorResponse = err.into();
        assert_eq!(expected, actual);
    }

    #[test]
    fn conversion_from_database_error() {
        let err = gejdr_core::sqlx::Error::PoolClosed;
        let expected = ErrorResponse {
            code: 500,
            message: "Internal database error".into(),
            ..Default::default()
        };
        let actual: ErrorResponse = err.into();
        assert_eq!(expected, actual);
    }
}
